package be.iminds.ilabt.jfed.testing.base;

import be.iminds.ilabt.jfed.call_log_output.HtmlLogOutput;
import be.iminds.ilabt.jfed.call_log_output.LogOutput;
import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.TestbedInfoSource;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticAggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticUserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.TargetAuthority;
import org.jetbrains.annotations.Contract;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.*;
import java.util.stream.Collectors;

/**
 * ApiTest:
 * <p>
 * for a test class you need:
 * - user info
 * - SA info if applicable
 * - tested API info
 * - connection (auto)
 * - logger (auto)
 * <p>
 * per test meta data
 * - name (default: auto from method name)
 * - dependencies and preferred grouping (name or methodName)
 * - hard dependency (skip if other failed or skipped)
 * - soft dependency (always run after other (no matter if it succeeds, skips or fails))
 * - group (a name that can be the same for a number of tests. These will be kept together, unless the other dependencies are used)
 * <p>
 * for each test log:
 * - commands and replies
 * - details, notes, warnings, errors and fatal errors (+ skips: the test can start to run and then decide it needs to be skipped)
 * - info on the test
 * - description of what it does
 * - description of what is expected
 */
public abstract class ApiTest<SpecificApiTestConfig extends ApiTestConfig> {
    private final static org.slf4j.Logger LOG = LoggerFactory.getLogger(ApiTest.class);

    @Nonnull protected final Logger logger;
    @Nonnull protected final TargetAuthority testedAuthority;
    @Nullable protected final GeniUserProvider geniUserProvider;
    @Nullable protected final GeniUser user;

    @Nonnull protected final JFedConnectionProvider connectionProvider;
    @Nonnull protected final TestbedInfoSource testbedInfoSource;
    @Nonnull protected final AuthorityFinder authorityFinder;
    @Nonnull protected final JFedPreferences jFedPreferences;


    @Nonnull public Logger getLogger() {
        return logger;
    }

    @Nonnull public TargetAuthority getTestedAuthority() {
        return testedAuthority;
    }

    @Nullable public GeniUser getUser() {
        return user;
    }

    @Nonnull public JFedConnectionProvider getConnectionProvider() {
        return connectionProvider;
    }

    @Nonnull public TestbedInfoSource getTestbedInfoSource() {
        return testbedInfoSource;
    }

    public AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory getAutomaticUserAndSliceApiWrapperFactory() {
        return new AutomaticUserAndSliceApiWrapper.BasicAutomaticUserAndSliceApiWrapperFactory(logger, geniUserProvider, connectionProvider, jFedPreferences);
    }

    public AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory getAutomaticAggregateManagerWrapperFactory() {
        return new AutomaticAggregateManagerWrapper.BasicAutomaticAggregateManagerWrapperFactory(logger, geniUserProvider, connectionProvider, jFedPreferences);
    }


    public ApiTest(@Nonnull Logger logger,
                   @Nonnull TargetAuthority testedAuthority,
                   @Nullable GeniUserProvider geniUserProvider,
                   @Nonnull JFedConnectionProvider connectionProvider,
                   @Nonnull TestbedInfoSource testbedInfoSource,
                   @Nonnull JFedPreferences jFedPreferences) {
        this(logger, testedAuthority, geniUserProvider, connectionProvider, testbedInfoSource,
                new AuthorityFinder(testbedInfoSource, jFedPreferences) /*safe fallback as AuthorityFinder does not keep state*/,
                jFedPreferences);


    }

    public ApiTest(@Nonnull Logger logger,
                   @Nonnull TargetAuthority testedAuthority,
                   @Nullable GeniUserProvider geniUserProvider,
                   @Nonnull JFedConnectionProvider connectionProvider,
                   @Nonnull TestbedInfoSource testbedInfoSource,
                   @Nonnull AuthorityFinder authorityFinder,
                   @Nonnull JFedPreferences jFedPreferences) {
        this.logger = logger;
        this.testedAuthority = testedAuthority;
        this.geniUserProvider = geniUserProvider;
        if (geniUserProvider != null)
            this.user = geniUserProvider.getLoggedInGeniUser();
        else
            this.user = null;
        this.connectionProvider = connectionProvider;
        this.testbedInfoSource = testbedInfoSource;
        this.authorityFinder = authorityFinder;
        this.jFedPreferences = jFedPreferences;
    }

    private SpecificApiTestConfig testConfig;

    public SpecificApiTestConfig getTestConfig() {
        return testConfig;
    }

    void setTestConfig(SpecificApiTestConfig testConfig) {
        this.testConfig = testConfig;
    }

    @Nonnull
    public static ApiTestMetaData getMetaData(@Nonnull ApiTest test) {
        return getMetaData(test.getClass());
    }
    @Nonnull
    public static ApiTestMetaData getMetaData(@Nonnull Class<? extends ApiTest> testClass) {
        try {
            Method m = testClass.getDeclaredMethod("getMetaData");
            ApiTestMetaData res = (ApiTestMetaData) m.invoke(null);
            if (res == null) {
                throw new NullPointerException("getMetaData() is incorrectly implemented. " +
                        "Each inheritor of ApiTest must have a static getMetaData() method, " +
                        "and that method should NOT return null.");
            }
            return res;
        } catch (NoSuchMethodException | IllegalAccessException | InvocationTargetException e) {
            throw new RuntimeException("getMetaData() is incorrectly implemented. " +
                    "Each inheritor of ApiTest must have a static getMetaData() method.", e);
        }
    }

    public abstract void setUp() throws Exception;

    protected ApiTestResult.ApiTestMethodResult currentTestResult;

    void setCurrentTestResult(ApiTestResult.ApiTestMethodResult currentTestResult) {
        this.currentTestResult = currentTestResult;
    }

    ApiTestResult.ApiTestMethodResult getCurrentTestResult() {
        return currentTestResult;
    }

    public LogOutput.TestResultState getCurrentResultState() {
        return currentTestResult.getState();
    }

    public JFedPreferences getJFedPreferences() {
        return jFedPreferences;
    }

    @java.lang.annotation.Retention(java.lang.annotation.RetentionPolicy.RUNTIME)
    @java.lang.annotation.Target({java.lang.annotation.ElementType.METHOD})
    public @interface Test {
        java.lang.String description() default "";

        java.lang.String[] groups() default {};

        java.lang.String[] hardDepends() default {};

        java.lang.String[] softDepends() default {};

        String dataProvider() default ""; //TODO implement

        //this concept (isReturnValue annotation) is under revision
//        /**
//         * If true, only the success of this test will be used as return value for the entire test run
//         * If there are no tests with this option, the success of all tests is used.
//         * If there are multiple tests in a class with this option, the result is unspecified!
//         * */
//        boolean ignoreForReturnValue() default false;
    }

    public @interface DataProvider { //TODO implement
        String name() default "";
    }


    @Deprecated
    //use currentTestResult.addExtraResult directly instead
    public void addExtraResult(@Nonnull Object val, @Nonnull String ... nestedkey) {
        assert nestedkey.length > 0;
        currentTestResult.addExtraResult(
                new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                        .setKey(nestedkey)
                        .setValue(val)
                        .setLarge(false));

        //for backward compatibility
        String usedKey = Arrays.asList(nestedkey).stream()
                .map(String::trim)
                .collect(Collectors.joining("_"))
                .toUpperCase().replaceAll(" ", "_");
        assert !usedKey.isEmpty();
        this.note(HtmlLogOutput.MACHINE_LOGLINE_PREFIX + " BEGIN_" + usedKey + " " + (val+"") + " END_" + usedKey);
    }

    public void note(String val) {
        currentTestResult.addLogLine(LogOutput.LogLineType.NOTE, val);
    }

    public void note(String val, boolean formatAsCode) {
        currentTestResult.addLogLine(LogOutput.LogLineType.NOTE, val, formatAsCode);
    }

    public void note(String val, Throwable t) {
        currentTestResult.addLogLine(LogOutput.LogLineType.NOTE, val, t);
    }

    public void note(String val, Throwable t, boolean formatAsCode) {
        currentTestResult.addLogLine(LogOutput.LogLineType.NOTE, val, t, formatAsCode);
    }

    public void skip(String reason) {
        currentTestResult.state = ApiTestResult.TestResultState.SKIPPED;
//        currentTestResult.addLogLine(ApiTestResult.MethodLogLineType.NOTE, "Test skipped: "+reason);
        throw new AutomatedTestRunner.SkipTestException(reason);
    }

    public void skip(String reason, Throwable t) {
        currentTestResult.state = ApiTestResult.TestResultState.SKIPPED;
        throw new AutomatedTestRunner.SkipTestException(reason, t);
    }

    //I don't like this, chaning if errors are fatal can lead to confusion

    @Deprecated private boolean errorsFatal = true;
    @Deprecated private boolean errorsAreWarnings = false;

    /**
     * errors will no longer be fatal, until setErrorsFatal() is called, or the current annotated test is done (so next test methods wil have fatal errors again).
     */
    @Deprecated
    public void setErrorsNotFatal() {
        errorsFatal = false;
        errorsAreWarnings = false;
    }

    @Deprecated
    public void setErrorsFatal() {
        errorsFatal = true;
        errorsAreWarnings = false;
    }

    /**
     * Errors are no longer errors, and are not fatal. They are warnings instead.
     * This works until setErrorsAreWarnings(false), setErrorsNotFatal() or setErrorsFatal() is called, or until the current annotated test is done (so next test methods wil have fatal errors again).
     *
     * using this is not advised
     */
    @Deprecated public void setErrorsAreWarnings(boolean warn) {
        errorsAreWarnings = warn;
    }


    /**
     * Makes the test fail, but continues executing the test code.
     *
     * @param val text describing the error
     */
    public void errorNonFatal(String val) {
        currentTestResult.state = ApiTestResult.TestResultState.FAILED;
        currentTestResult.addLogLine(LogOutput.LogLineType.ERROR, val);
    }

    public void errorNonFatal(String val, Throwable t) {
        currentTestResult.state = ApiTestResult.TestResultState.FAILED;
        currentTestResult.addLogLine(LogOutput.LogLineType.ERROR, val, t);
    }

    @Contract("_ -> fail")
    public void errorFatal(String val) {
        fatalError(val);
    }

    @Contract("_,_ -> fail")
    public void errorFatal(String val, Throwable t) {
        fatalError(val, t);
    }

    @Contract("_ -> fail")
    public void fatalError(String val) {
        fatalError(val, null);
    }

    @Contract("_,_ -> fail")
    public void fatalError(String val, Throwable t) {
        if (errorsAreWarnings) {
            warn(val, t);
        } else {
            if (errorsFatal) {
//        currentTestResult.addLogLine(ApiTestResult.MethodLogLineType.FATAL_ERROR, val);
                currentTestResult.state = ApiTestResult.TestResultState.FAILED;
                throw new AssertionError(val, t);
            } else {
                errorNonFatal(val, t);
            }
        }
    }

    public void warn(String val, boolean formatAsCode) {
        currentTestResult.addLogLine(LogOutput.LogLineType.WARN, val, formatAsCode);
        if (currentTestResult.state == ApiTestResult.TestResultState.SUCCESS)
            currentTestResult.state = ApiTestResult.TestResultState.WARN;
    }

    public void warn(String val) {
        this.warn(val, false/*formatAsCode*/);
    }

    public void warn(String val, Throwable e, boolean formatAsCode) {
        currentTestResult.addLogLine(LogOutput.LogLineType.WARN, val, e);
        if (currentTestResult.state == ApiTestResult.TestResultState.SUCCESS)
            currentTestResult.state = ApiTestResult.TestResultState.WARN;
    }

    public void warn(String val, Throwable e) {
        this.warn(val, e, false/*formatAsCode*/);
    }

    public void warnIfNot(boolean v) {
        warnIfNot(v, "warnIfNot failed");
    }

    public void warnIfNot(boolean v, String failText) {
        if (!v)
            warn(failText);
    }

    public void errorNonFatalIfNot(boolean v) {
        errorNonFatalIfNot(v, "warnIfNot failed");
    }

    public void errorNonFatalIfNot(boolean v, String failText) {
        if (!v)
            errorNonFatal(failText);
    }

    @Contract("false -> fail")
    public void assertTrue(boolean v) {
        assertTrue(v, "assertTrue failed");
    }

    @Contract("false,_ -> fail")
    public void assertTrue(boolean v, String failText) {
        if (!v)
            fatalError(failText);
    }


    @Contract("true -> fail")
    public void assertFalse(boolean v) {
        assertFalse(v, "assertFalse failed");
    }

    @Contract("true,_ -> fail")
    public void assertFalse(boolean v, String failText) {
        if (v)
            fatalError(failText);
    }


    @Contract("null -> fail")
    public void assertNotNull(Object o) {
        assertNotNull(o, "object is null");
    }

    @Contract("null,_ -> fail")
    public void assertNotNull(Object o, String failText) {
        if (o == null)
            fatalError(failText);
    }


    public void assertNonEmptyString(String s) {
        assertNonEmptyString(s, "string should not be empty");
    }

    public void assertNonEmptyString(String s, String failText) {
        if (s == null) {
            fatalError(failText);
        } else {
            if (s.trim().isEmpty())
                fatalError(failText);
        }
    }

    public void assertNotEmpty(Collection c) {
        assertNotEmpty(c, "list is empty");
    }

    public void assertEmpty(Collection c) {
        assertEmpty(c, "list is not empty");
    }

    public void assertNotEmpty(Collection c, String failText) {
        if (c.isEmpty())
            fatalError(failText);
    }

    public void assertEmpty(Collection c, String failText) {
        if (!c.isEmpty())
            fatalError(failText);
    }


    @Contract("!null -> fail")
    public void assertNull(Object o) {
        assertNotNull(o, "object is null");
    }

    @Contract("!null,_ -> fail")
    public void assertNull(Object o, String failText) {
        if (o != null)
            fatalError(failText);
    }

    public void assertEquals(Object a, Object b, String failText) {
        if (!Objects.equals(a, b))
            fatalError(failText + " -> " + (a instanceof String
                    ? "object are not equal: a=\"" + a.toString() + "\" b=\"" +
                    b.toString() + "\"" : "object are not equal: a=" + a.toString() + "b=" + b.toString()));
    }

    public void assertEquals(Object a, Object b) {
        assertEquals(a, b, a instanceof String ? "object are not equal: a=\"" + a.toString() + "\" b=\""
                + b.toString() + "\"" : "object are not equal: a=" + a.toString() + "b=" + b.toString());
    }

    /**
     * assert that 2 dates are equal, to second precision. sub seconds precision may differ
     */
    public void assertEqualDatesNoMs(Date a, Date b) {
        long as = a.getTime() / 1000L;
        long bs = b.getTime() / 1000L;
        LOG.debug("assertEqualDatesNoMs compare as=" + as + " bs=" + bs + "   a=" + a + "  b=" + b);
        if (as != bs)
            fatalError("dates are not equal: " + a.toString() + " differs from " + b.toString() + "");
    }


    public void assertNotEquals(Object a, Object b, String failText) {
        if (Objects.equals(a, b))
            fatalError(failText + " -> object are equal: a=b=" + a.toString());
    }

    public void assertNotEquals(Object a, Object b) {
        assertNotEquals(a, b, "object are equal: a=b=" + a.toString());
    }

    public <T> T assertInstanceOf(Object o, Class<T> c) {
        return assertInstanceOf(o, c, "object is not an instance of class \"" + c.getName() + "\". (It is instance of class " + c.getName() + ")");
    }

    public <T> T assertInstanceOf(Object o, Class<T> c, String failText) {
        if (!c.isInstance(o)) {
            fatalError(failText + " -> object is not an instance of class \"" + c.getName() + "\". (It is instance of class " + c.getName() + ")");
            return null;
        } else
            return c.cast(o);
    }


    @Nullable
    public List assertMapContainsList(@Nonnull Map t, @Nonnull String key) {
        Object o = t.get(key);
        assertNotNull(o, "Map does not contain key \"" + key + "\"");
        if (o == null) return null;
        assertTrue(o instanceof Object[], "value for " + key + " is not a List but a " + o.getClass().getName());
        if ((o instanceof Object[])) {
            return Arrays.asList((Object[]) o);
        } else
            return null;
    }

    @SuppressWarnings("unchecked")
    public <T> List<T> assertMapContainsList(Map t, String key, Class<T> clazz) {
        List list = assertMapContainsList(t, key);
        for (Object o : list) {
            assertInstanceOf(o, clazz,
                    "Value in List was expected to be a " + clazz.getName() + ", but got a " + o.getClass().getName());
        }
        return (List<T>) list;
    }

    public Map assertMapContainsMap(Map t, String key) {
        Object o = t.get(key);
        assertNotNull(o, "Map does not contain key \"" + key + "\"");
        if (o == null) return null;
        assertInstanceOf(o, Map.class, "value for " + key + " is not a Map but a " + o.getClass().getName());
        if (!(o instanceof Map)) return null;
        return (Map) o;
    }

    public String assertMapContainsString(Map t, String key) {
        Object o = t.get(key);
        assertNotNull(o, "Map does not contain key \"" + key + "\"");
        if (o == null) return null;
        assertEquals(o.getClass(), String.class, "value for " + key + " is not a String but a " + o.getClass().getName());
        if (!Objects.equals(o.getClass(), String.class)) return null;
        return (String) o;
    }

    public String assertMapContainsString(Map t, String key, String expectedValue) {
        Object o = t.get(key);
        assertNotNull(o, "Map does not contain key \"" + key + "\"");
        if (o == null) return null;
        assertEquals(o.getClass(), String.class, "value for " + key + " is not a String but a " + o.getClass().getName());
        assertEquals(o, expectedValue, "value for " + key + " is not \"" + expectedValue + "\" but \"" + o + "\"");
        if (!Objects.equals(o.getClass(), String.class)) return null;
        return (String) o;
    }

    public Object assertMapContainsObject(Map t, String key) {
        Object o = t.get(key);
        assertNotNull(o, "Map does not contain key \"" + key + "\"");
        return o;
    }

    public String assertMapContainsNonemptyString(Map t, String key) {
        Object o = t.get(key);
        assertNotNull(o, "Map does not contain key \"" + key + "\"");
        if (o == null) return null;
        assertEquals(o.getClass(), String.class, "value for " + key + " is not a String but a " + o.getClass().getName());
        if (!Objects.equals(o.getClass(), String.class)) return null;
        String s = (String) o;
        assertTrue(s.length() > 0, "value for " + key + " is an empty string");
        return s;
    }


    @Nullable
    public Map<String, Object> assertStringObjectMap(Object o) {
        Map t =
                assertInstanceOf(o, Map.class, "Expected a dictionary (=Map), but is of type " + o.getClass().getName());
        if (t != null) {
            for (Object key : t.keySet()) {
                assertInstanceOf(key, String.class, "Expected dictionary to only contain Strings as keys, but got a " + key.getClass().getName());
            }

            //noinspection unchecked
            return (Map<String, Object>) t;
        } else
            return null;
    }

    /**
     * @param type is allowed to be null if no type to test
     */
    public void assertValidUrn(String urn, String type) {
        assertValidUrn(urn, type, null);

    }

    public void assertValidUrn(String urn, String type, String message) {
        String msgHelp = message == null ? "" : message + ": ";
        GeniUrn geniUrn = GeniUrn.parse(urn);
        assertTrue(geniUrn != null, msgHelp + "Urn is not valid: \"" + urn + "\"");
        if (geniUrn == null) return;
        if (type != null)
            assertEquals(geniUrn.getEncodedResourceType(), type, msgHelp + "Urn type is not correct in urn=\"" + urn + "\"");
        assertFalse(geniUrn.getEncodedTopLevelAuthority().isEmpty(), msgHelp + "Urn authority is empty in urn=\"" + urn + "\"");
        assertFalse(geniUrn.getEncodedResourceName().isEmpty(), msgHelp + "Urn resource name is empty in urn=\"" + urn + "\"");
    }

}

